@@ -5,7 +5,8 @@ use crossbeam_channel as channel;
 use ibc::ics02_client::client_consensus::{AnyConsensusState, AnyConsensusStateWithHeight};
 use ibc::ics02_client::client_state::{AnyClientState, IdentifiedAnyClientState};
 use ibc::ics02_client::events::UpdateClient;
-use ibc::ics02_client::misbehaviour::AnyMisbehaviour;
+use ibc::ics02_client::misbehaviour::MisbehaviourEvidence;
+use ibc::ics03_connection::connection::IdentifiedConnectionEnd;
 use ibc::ics04_channel::channel::IdentifiedChannelEnd;
 use ibc::ics04_channel::packet::{PacketMsgType, Sequence};
 use ibc::query::QueryTxRequest;
@@ -32,6 +33,7 @@ use ibc_proto::ibc::core::channel::v1::{
 use ibc_proto::ibc::core::client::v1::{QueryClientStatesRequest, QueryConsensusStatesRequest};
 use ibc_proto::ibc::core::commitment::v1::MerkleProof;
 use ibc_proto::ibc::core::connection::v1::QueryClientConnectionsRequest;
+use ibc_proto::ibc::core::connection::v1::QueryConnectionsRequest;
 
 use crate::{
     connection::ConnectionMsgType,
@@ -190,6 +192,13 @@ impl ChainHandle for ProdChainHandle {
         })
     }
 
+    fn query_connections(
+        &self,
+        request: QueryConnectionsRequest,
+    ) -> Result<Vec<IdentifiedConnectionEnd>, Error> {
+        self.send(|reply_to| ChainRequest::QueryConnections { request, reply_to })
+    }
+
     fn query_connection_channels(
         &self,
         request: QueryConnectionChannelsRequest,
@@ -275,7 +284,7 @@ impl ChainHandle for ProdChainHandle {
         trusted_height: Height,
         target_height: Height,
         client_state: AnyClientState,
-    ) -> Result<AnyHeader, Error> {
+    ) -> Result<(AnyHeader, Vec<AnyHeader>), Error> {
         self.send(|reply_to| ChainRequest::BuildHeader {
             trusted_height,
             target_height,
@@ -306,7 +315,7 @@ impl ChainHandle for ProdChainHandle {
         &self,
         update_event: UpdateClient,
         client_state: AnyClientState,
-    ) -> Result<Option<AnyMisbehaviour>, Error> {
+    ) -> Result<Option<MisbehaviourEvidence>, Error> {
         self.send(|reply_to| ChainRequest::BuildMisbehaviour {
             client_state,
             update_event,
