@@ -12,7 +12,7 @@ use ibc::downcast;
 use ibc::events::IbcEvent;
 use ibc::ics02_client::client_consensus::{AnyConsensusState, AnyConsensusStateWithHeight};
 use ibc::ics02_client::client_state::{AnyClientState, IdentifiedAnyClientState};
-use ibc::ics03_connection::connection::ConnectionEnd;
+use ibc::ics03_connection::connection::{ConnectionEnd, IdentifiedConnectionEnd};
 use ibc::ics04_channel::channel::{ChannelEnd, IdentifiedChannelEnd};
 use ibc::ics04_channel::packet::{PacketMsgType, Sequence};
 use ibc::ics07_tendermint::client_state::{AllowUpdate, ClientState as TendermintClientState};
@@ -44,6 +44,7 @@ use crate::config::ChainConfig;
 use crate::error::{Error, Kind};
 use crate::event::monitor::{EventReceiver, EventSender};
 use crate::keyring::{KeyEntry, KeyRing};
+use crate::light_client::Verified;
 use crate::light_client::{mock::LightClient as MockLightClient, LightClient};
 
 /// The representation of a mocked chain as the relayer sees it.
@@ -177,7 +178,7 @@ impl Chain for MockChain {
     fn query_connections(
         &self,
         _request: QueryConnectionsRequest,
-    ) -> Result<Vec<ConnectionId>, Error> {
+    ) -> Result<Vec<IdentifiedConnectionEnd>, Error> {
         unimplemented!()
     }
 
@@ -325,15 +326,33 @@ impl Chain for MockChain {
     fn build_header(
         &self,
         trusted_height: Height,
-        trusted_light_block: Self::LightBlock,
-        target_light_block: Self::LightBlock,
-    ) -> Result<Self::Header, Error> {
-        Ok(Self::Header {
-            signed_header: target_light_block.signed_header.clone(),
-            validator_set: target_light_block.validators,
+        target_height: Height,
+        client_state: &AnyClientState,
+        light_client: &mut dyn LightClient<Self>,
+    ) -> Result<(Self::Header, Vec<Self::Header>), Error> {
+        let succ_trusted = light_client.fetch(trusted_height.increment())?;
+
+        let Verified { target, supporting } =
+            light_client.verify(trusted_height, target_height, client_state)?;
+
+        let target_header = Self::Header {
+            signed_header: target.signed_header,
+            validator_set: target.validators,
             trusted_height,
-            trusted_validator_set: trusted_light_block.validators,
-        })
+            trusted_validator_set: succ_trusted.validators.clone(),
+        };
+
+        let supporting_headers = supporting
+            .into_iter()
+            .map(|h| Self::Header {
+                signed_header: h.signed_header,
+                validator_set: h.validators,
+                trusted_height,
+                trusted_validator_set: succ_trusted.validators.clone(),
+            })
+            .collect();
+
+        Ok((target_header, supporting_headers))
     }
 
     fn query_consensus_states(
@@ -371,7 +390,7 @@ pub mod test_utils {
 
     use ibc::ics24_host::identifier::ChainId;
 
-    use crate::config::ChainConfig;
+    use crate::config::{ChainConfig, GasPrice};
 
     /// Returns a very minimal chain configuration, to be used in initializing `MockChain`s.
     pub fn get_basic_chain_config(id: &str) -> ChainConfig {
@@ -384,9 +403,9 @@ pub mod test_utils {
             account_prefix: "".to_string(),
             key_name: "".to_string(),
             store_prefix: "".to_string(),
-            gas: None,
-            fee_denom: "stake".to_string(),
-            fee_amount: Some(1000),
+            max_gas: None,
+            gas_price: GasPrice::new(0.001, "uatom".to_string()),
+            gas_adjustment: None,
             max_msg_num: None,
             max_tx_size: None,
             clock_drift: Duration::from_secs(5),
