@@ -1,8 +1,13 @@
 use std::time::Duration;
 
+use crate::chain::counterparty::connection_state_on_destination;
+use crate::util::retry::RetryResult;
+use anomaly::BoxError;
+use ibc_proto::ibc::core::connection::v1::QueryConnectionsRequest;
 use prost_types::Any;
 use serde::Serialize;
 use thiserror::Error;
+use tracing::debug;
 use tracing::{error, warn};
 
 use ibc::events::IbcEvent;
@@ -17,11 +22,11 @@ use ibc::ics03_connection::msgs::conn_open_try::MsgConnectionOpenTry;
 use ibc::ics24_host::identifier::{ChainId, ClientId, ConnectionId};
 use ibc::timestamp::ZERO_DURATION;
 use ibc::tx_msg::Msg;
-use ibc::Height as ICSHeight;
 
 use crate::chain::handle::ChainHandle;
 use crate::error::Error;
 use crate::foreign_client::{ForeignClient, ForeignClientError};
+use crate::object::Connection as WorkerConnectionObject;
 
 /// Maximum value allowed for packet delay on any new connection that the relayer establishes.
 pub const MAX_PACKET_DELAY: Duration = Duration::from_secs(120);
@@ -36,6 +41,12 @@ pub enum ConnectionError {
     #[error("connection constructor error: {0}")]
     ConstructorFailed(String),
 
+    #[error("failed due to missing local channel id")]
+    MissingLocalConnectionId,
+
+    #[error("failed due to missing counterparty connection id ")]
+    MissingCounterpartyConnectionId,
+
     #[error("failed during a query to chain id {0} due to underlying error: {1}")]
     QueryError(ChainId, Error),
 
@@ -52,14 +63,14 @@ pub enum ConnectionError {
 pub struct ConnectionSide {
     pub(crate) chain: Box<dyn ChainHandle>,
     client_id: ClientId,
-    connection_id: ConnectionId,
+    connection_id: Option<ConnectionId>,
 }
 
 impl ConnectionSide {
     pub fn new(
         chain: Box<dyn ChainHandle>,
         client_id: ClientId,
-        connection_id: ConnectionId,
+        connection_id: Option<ConnectionId>,
     ) -> Self {
         Self {
             chain,
@@ -67,6 +78,9 @@ impl ConnectionSide {
             connection_id,
         }
     }
+    pub fn connection_id(&self) -> Option<&ConnectionId> {
+        self.connection_id.as_ref()
+    }
 }
 
 impl Serialize for ConnectionSide {
@@ -77,7 +91,7 @@ impl Serialize for ConnectionSide {
         #[derive(Debug, Serialize)]
         struct ConnectionSide<'a> {
             client_id: &'a ClientId,
-            connection_id: &'a ConnectionId,
+            connection_id: &'a Option<ConnectionId>,
         }
 
         let value = ConnectionSide {
@@ -132,6 +146,92 @@ impl Connection {
 
         Ok(c)
     }
+    pub fn restore_from_event(
+        chain: Box<dyn ChainHandle>,
+        counterparty_chain: Box<dyn ChainHandle>,
+        connection_open_event: IbcEvent,
+    ) -> Result<Connection, BoxError> {
+        let connection_event_attributes = connection_open_event
+            .connection_attributes()
+            .ok_or_else(|| {
+                ConnectionError::Failed(format!(
+                    "a connection object cannot be built from {}",
+                    connection_open_event
+                ))
+            })?;
+
+        let connection_id = connection_event_attributes.connection_id.clone();
+
+        let counterparty_connection_id = connection_event_attributes
+            .counterparty_connection_id
+            .clone();
+
+        let client_id = connection_event_attributes.client_id.clone();
+        let counterparty_client_id = connection_event_attributes.counterparty_client_id.clone();
+
+        Ok(Connection {
+            // The event does not include the connection delay.
+            delay_period: Default::default(),
+            a_side: ConnectionSide::new(chain, client_id, connection_id),
+            b_side: ConnectionSide::new(
+                counterparty_chain,
+                counterparty_client_id,
+                counterparty_connection_id,
+            ),
+        })
+    }
+
+    /// Recreates a 'Connection' object from the worker's object built from chain state scanning.
+    /// The connection must exist on chain.
+    pub fn restore_from_state(
+        chain: Box<dyn ChainHandle>,
+        counterparty_chain: Box<dyn ChainHandle>,
+        connection: WorkerConnectionObject,
+        height: Height,
+    ) -> Result<(Connection, State), BoxError> {
+        let a_connection = chain.query_connection(&connection.src_connection_id, height)?;
+        let client_id = a_connection.client_id();
+        let delay_period = a_connection.delay_period();
+
+        let counterparty_connection_id = a_connection.counterparty().connection_id.clone();
+
+        let counterparty_client_id = a_connection.counterparty().client_id();
+
+        let mut handshake_connection = Connection {
+            delay_period,
+            a_side: ConnectionSide::new(
+                chain.clone(),
+                client_id.clone(),
+                Some(connection.src_connection_id.clone()),
+            ),
+            b_side: ConnectionSide::new(
+                counterparty_chain.clone(),
+                counterparty_client_id.clone(),
+                counterparty_connection_id.clone(),
+            ),
+        };
+
+        if a_connection.state_matches(&State::Init) && counterparty_connection_id.is_none() {
+            let req = QueryConnectionsRequest {
+                pagination: ibc_proto::cosmos::base::query::pagination::all(),
+            };
+            let connections: Vec<IdentifiedConnectionEnd> =
+                counterparty_chain.query_connections(req)?;
+
+            for conn in connections {
+                if let Some(remote_connection_id) =
+                    conn.connection_end.counterparty().connection_id()
+                {
+                    if remote_connection_id == &connection.src_connection_id {
+                        handshake_connection.b_side.connection_id = Some(conn.connection_id);
+                        break;
+                    }
+                }
+            }
+        }
+
+        Ok((handshake_connection, *a_connection.state()))
+    }
 
     pub fn find(
         a_client: ForeignClient,
@@ -176,12 +276,12 @@ impl Connection {
             a_side: ConnectionSide {
                 chain: a_client.dst_chain.clone(),
                 client_id: a_client.id.clone(),
-                connection_id: conn_end_a.id().clone(),
+                connection_id: Some(conn_end_a.id().clone()),
             },
             b_side: ConnectionSide {
                 chain: b_client.dst_chain.clone(),
                 client_id: b_client.id.clone(),
-                connection_id: b_conn_id,
+                connection_id: Some(b_conn_id),
             },
         };
 
@@ -228,12 +328,12 @@ impl Connection {
         &self.b_side.client_id
     }
 
-    pub fn src_connection_id(&self) -> &ConnectionId {
-        &self.a_side.connection_id
+    pub fn src_connection_id(&self) -> Option<&ConnectionId> {
+        self.a_side.connection_id()
     }
 
-    pub fn dst_connection_id(&self) -> &ConnectionId {
-        &self.b_side.connection_id
+    pub fn dst_connection_id(&self) -> Option<&ConnectionId> {
+        self.b_side.connection_id()
     }
 
     pub fn flipped(&self) -> Connection {
@@ -261,7 +361,7 @@ impl Connection {
                     continue;
                 }
                 Ok(result) => {
-                    self.a_side.connection_id = extract_connection_id(&result)?.clone();
+                    self.a_side.connection_id = Some(extract_connection_id(&result)?.clone());
                     println!("🥂  {} => {:#?}\n", self.a_side.chain.id(), result);
                     break;
                 }
@@ -278,7 +378,7 @@ impl Connection {
                     continue;
                 }
                 Ok(result) => {
-                    self.b_side.connection_id = extract_connection_id(&result)?.clone();
+                    self.b_side.connection_id = Some(extract_connection_id(&result)?.clone());
                     println!("{}  {} => {:#?}\n", done, self.b_side.chain.id(), result);
                     break;
                 }
@@ -289,20 +389,24 @@ impl Connection {
         while counter < MAX_RETRIES {
             counter += 1;
 
+            let src_connection_id = self
+                .src_connection_id()
+                .ok_or(ConnectionError::MissingLocalConnectionId)?;
+            let dst_connection_id = self
+                .dst_connection_id()
+                .ok_or(ConnectionError::MissingCounterpartyConnectionId)?;
+
             // Continue loop if query error
-            let a_connection = a_chain.query_connection(&self.src_connection_id(), Height::zero());
+            let a_connection = a_chain.query_connection(&src_connection_id, Height::zero());
             if a_connection.is_err() {
                 continue;
             }
-            let b_connection = b_chain.query_connection(&self.dst_connection_id(), Height::zero());
+            let b_connection = b_chain.query_connection(&dst_connection_id, Height::zero());
             if b_connection.is_err() {
                 continue;
             }
 
-            match (
-                a_connection.unwrap().state().clone(),
-                b_connection.unwrap().state().clone(),
-            ) {
+            match (a_connection.unwrap().state(), b_connection.unwrap().state()) {
                 (State::Init, State::TryOpen) | (State::TryOpen, State::TryOpen) => {
                     // Ack to a_chain
                     match self.flipped().build_conn_ack_and_send() {
@@ -347,6 +451,73 @@ impl Connection {
         )))
     }
 
+    pub fn counterparty_state(&self) -> Result<State, ConnectionError> {
+        // Source connection ID must be specified
+        let connection_id = self
+            .src_connection_id()
+            .ok_or(ConnectionError::MissingLocalConnectionId)?;
+
+        let connection_end = self
+            .src_chain()
+            .query_connection(&connection_id, Height::zero())
+            .map_err(|_| {
+                ConnectionError::Failed(format!(
+                    "failed to query the connection for {}",
+                    connection_id
+                ))
+            })?;
+
+        let connection = IdentifiedConnectionEnd {
+            connection_end,
+            connection_id: connection_id.clone(),
+        };
+
+        connection_state_on_destination(connection, self.dst_chain().as_ref()).map_err(|e| {
+            ConnectionError::Failed(format!(
+                "failed to query the connection state on destination for {}; caused by {}",
+                connection_id, e
+            ))
+        })
+    }
+
+    pub fn handshake_step(&mut self, state: State) -> Result<Vec<IbcEvent>, ConnectionError> {
+        match (state, self.counterparty_state()?) {
+            (State::Init, State::Uninitialized) => Ok(vec![self.build_conn_try_and_send()?]),
+            (State::Init, State::Init) => Ok(vec![self.build_conn_try_and_send()?]),
+            (State::TryOpen, State::Init) => Ok(vec![self.build_conn_ack_and_send()?]),
+            (State::TryOpen, State::TryOpen) => Ok(vec![self.build_conn_ack_and_send()?]),
+            (State::Open, State::TryOpen) => Ok(vec![self.build_conn_confirm_and_send()?]),
+            _ => Ok(vec![]),
+        }
+    }
+
+    pub fn step_state(&mut self, state: State, index: u64) -> RetryResult<(), u64> {
+        let done = '🥳';
+
+        match self.handshake_step(state) {
+            Err(e) => {
+                error!("failed {:?} with error {}", state, e);
+                RetryResult::Retry(index)
+            }
+            Ok(ev) => {
+                debug!("{} => {:#?}\n", done, ev);
+                RetryResult::Ok(())
+            }
+        }
+    }
+
+    pub fn step_event(&mut self, event: IbcEvent, index: u64) -> RetryResult<(), u64> {
+        let state = match event {
+            IbcEvent::OpenInitConnection(_) => State::Init,
+            IbcEvent::OpenTryConnection(_) => State::TryOpen,
+            IbcEvent::OpenAckConnection(_) => State::Open,
+            IbcEvent::OpenConfirmConnection(_) => State::Open,
+            _ => State::Uninitialized,
+        };
+
+        self.step_state(state, index)
+    }
+
     /// Retrieves the connection from destination and compares against the expected connection
     /// built from the message type (`msg_type`) and options (`opts`).
     /// If the expected and the destination connections are compatible, it returns the expected connection
@@ -354,6 +525,10 @@ impl Connection {
         &self,
         msg_type: ConnectionMsgType,
     ) -> Result<ConnectionEnd, ConnectionError> {
+        let dst_connection_id = self
+            .dst_connection_id()
+            .ok_or(ConnectionError::MissingCounterpartyConnectionId)?;
+
         let prefix = self
             .src_chain()
             .query_commitment_prefix()
@@ -362,7 +537,7 @@ impl Connection {
         // If there is a connection present on the destination chain, it should look like this:
         let counterparty = Counterparty::new(
             self.src_client_id().clone(),
-            Option::from(self.src_connection_id().clone()),
+            self.src_connection_id().cloned(),
             prefix,
         );
 
@@ -389,21 +564,21 @@ impl Connection {
         // Retrieve existing connection if any
         let dst_connection = self
             .dst_chain()
-            .query_connection(self.dst_connection_id(), ICSHeight::default())
+            .query_connection(dst_connection_id, Height::zero())
             .map_err(|e| ConnectionError::QueryError(self.dst_chain().id(), e))?;
 
         // Check if a connection is expected to exist on destination chain
         // A connection must exist on destination chain for Ack and Confirm Tx-es to succeed
         if dst_connection.state_matches(&State::Uninitialized) {
             return Err(ConnectionError::Failed(format!(
-                "missing connection {} on source chain {}",
+                "missing connection {:?} on source chain {}",
                 self.src_connection_id().clone(),
                 self.dst_chain().id()
             )));
         }
 
         check_destination_connection_state(
-            self.dst_connection_id().clone(),
+            dst_connection_id.clone(),
             dst_connection,
             dst_expected_connection.clone(),
         )?;
@@ -492,9 +667,13 @@ impl Connection {
 
     /// Attempts to build a MsgConnOpenTry.
     pub fn build_conn_try(&self) -> Result<Vec<Any>, ConnectionError> {
+        let src_connection_id = self
+            .src_connection_id()
+            .ok_or(ConnectionError::MissingLocalConnectionId)?;
+
         let src_connection = self
             .src_chain()
-            .query_connection(self.src_connection_id(), ICSHeight::default())
+            .query_connection(src_connection_id, Height::zero())
             .map_err(|e| ConnectionError::QueryError(self.src_chain().id(), e))?;
 
         // TODO - check that the src connection is consistent with the try options
@@ -533,7 +712,7 @@ impl Connection {
             .src_chain()
             .build_connection_proofs_and_client_state(
                 ConnectionMsgType::OpenTry,
-                self.src_connection_id(),
+                src_connection_id,
                 self.src_client_id(),
                 query_height,
             )
@@ -568,14 +747,20 @@ impl Connection {
 
         let counterparty = Counterparty::new(
             self.src_client_id().clone(),
-            Option::from(self.src_connection_id().clone()),
+            self.src_connection_id().cloned(),
             prefix,
         );
 
+        let previous_connection_id = if src_connection.counterparty().connection_id.is_none() {
+            self.b_side.connection_id.clone()
+        } else {
+            src_connection.counterparty().connection_id.clone()
+        };
+
         let new_msg = MsgConnectionOpenTry {
             client_id: self.dst_client_id().clone(),
             client_state,
-            previous_connection_id: None,
+            previous_connection_id,
             counterparty,
             counterparty_versions,
             proofs,
@@ -617,6 +802,13 @@ impl Connection {
 
     /// Attempts to build a MsgConnOpenAck.
     pub fn build_conn_ack(&self) -> Result<Vec<Any>, ConnectionError> {
+        let src_connection_id = self
+            .src_connection_id()
+            .ok_or(ConnectionError::MissingLocalConnectionId)?;
+        let dst_connection_id = self
+            .dst_connection_id()
+            .ok_or(ConnectionError::MissingCounterpartyConnectionId)?;
+
         let _expected_dst_connection = self
             .validated_expected_connection(ConnectionMsgType::OpenAck)
             .map_err(|e|
@@ -626,7 +818,7 @@ impl Connection {
 
         let src_connection = self
             .src_chain()
-            .query_connection(self.src_connection_id(), ICSHeight::default())
+            .query_connection(src_connection_id, Height::zero())
             .map_err(|e| ConnectionError::QueryError(self.src_chain().id(), e))?;
 
         // TODO - check that the src connection is consistent with the ack options
@@ -650,7 +842,7 @@ impl Connection {
             .src_chain()
             .build_connection_proofs_and_client_state(
                 ConnectionMsgType::OpenAck,
-                self.src_connection_id(),
+                src_connection_id,
                 self.src_client_id(),
                 query_height,
             )
@@ -671,8 +863,8 @@ impl Connection {
         })?;
 
         let new_msg = MsgConnectionOpenAck {
-            connection_id: self.dst_connection_id().clone(),
-            counterparty_connection_id: self.src_connection_id().clone(),
+            connection_id: dst_connection_id.clone(),
+            counterparty_connection_id: src_connection_id.clone(),
             client_state,
             proofs,
             version: src_connection.versions()[0].clone(),
@@ -713,6 +905,13 @@ impl Connection {
 
     /// Attempts to build a MsgConnOpenConfirm.
     pub fn build_conn_confirm(&self) -> Result<Vec<Any>, ConnectionError> {
+        let src_connection_id = self
+            .src_connection_id()
+            .ok_or(ConnectionError::MissingLocalConnectionId)?;
+        let dst_connection_id = self
+            .dst_connection_id()
+            .ok_or(ConnectionError::MissingCounterpartyConnectionId)?;
+
         let _expected_dst_connection = self
             .validated_expected_connection(ConnectionMsgType::OpenAck)
             .map_err(|e| {
@@ -724,13 +923,14 @@ impl Connection {
             .src_chain()
             .query_latest_height()
             .map_err(|e| ConnectionError::QueryError(self.src_chain().id(), e))?;
+
         let _src_connection = self
             .src_chain()
-            .query_connection(self.src_connection_id(), query_height)
+            .query_connection(src_connection_id, query_height)
             .map_err(|_| {
                 ConnectionError::Failed(format!(
                     "missing connection {} on source chain",
-                    self.src_connection_id()
+                    src_connection_id
                 ))
             })?;
 
@@ -740,7 +940,7 @@ impl Connection {
             .src_chain()
             .build_connection_proofs_and_client_state(
                 ConnectionMsgType::OpenConfirm,
-                self.src_connection_id(),
+                src_connection_id,
                 self.src_client_id(),
                 query_height,
             )
@@ -761,7 +961,7 @@ impl Connection {
         })?;
 
         let new_msg = MsgConnectionOpenConfirm {
-            connection_id: self.dst_connection_id().clone(),
+            connection_id: dst_connection_id.clone(),
             proofs,
             signer,
         };
@@ -843,8 +1043,7 @@ fn check_destination_connection_state(
         && existing_connection.counterparty().client_id()
             == expected_connection.counterparty().client_id();
 
-    let good_state =
-        existing_connection.state().clone() as u32 <= expected_connection.state().clone() as u32;
+    let good_state = *existing_connection.state() as u32 <= *expected_connection.state() as u32;
 
     let good_connection_ids = existing_connection.counterparty().connection_id().is_none()
         || existing_connection.counterparty().connection_id()
